// File        : Fract.prg
// Description :
//   This file demonstrates a simple algorithm to
//   calculate and display the Mandelbrot set in text mode.
//
// Compile     : clipper fract /a/m/n/w
// Link        : blinker file fract
// Copyright   : Wynn Systems Development Group, Ltd.

#define MAX_DIST_SQR 16
#define MAX_ITERATIONS 12

#xtranslate Col2Real( <nCol> ) => ;
	( (<nCol> - (maxCol()/2) ) / (maxCol()/4) )

#xtranslate Row2Imag( <nRow> ) => ;
	( -(<nRow> - (maxRow()/2) ) / (maxRow()/4) )

static saColors := { "R+/W", "RB/W", "B/W", "BG/W", "G/W", ;
	"R/W", "RB+/W", "BG+/W", "GR+/W", "W+/W", "G+/W", "W/W" }


// Procedure   : Fract()
// Description : Draw the Mandelbrot fractal set in text mode.
// Parameters  :
//   None
//
// Details     :
//   This routine draws the Mandelbrot fractal set by mapping
//   each row and column pair into a point on the complex
//   plain.  This point is then tested for membership in the
//   Mandelbrot set.  It is then displayed on the screen, in a
//   color which is based on the number of iterations required
//   before divergence.

PROCEDURE Fract()

	local nRow
	local nCol
	local nIterCount
	local nMaxRow     := maxRow()
	local nMaxCol     := maxCol()
	local nColorCount := len( saColors )
	local nMidRow     := int( nMaxRow / 2 )

	// clear the screen
	scroll()

	@ 0, 0 say padc( "Clipper Fractal Drawing Program", maxCol() + 1 ) color "N/W"

	// process all rows
	for nRow := 1 to nMidRow

		// process all columns
		for nCol := 0 to nMaxCol

			// check for membership in the Mandelbrot set
			if ( ( nIterCount := IterCount( ;
					Col2Real( nCol ), ;
					Row2Imag( nRow ) ) ) < MAX_ITERATIONS )

				// diverged, display it in a color determined
				// by number of iterations before divergence
				@ nRow, nCol say "" color saColors[ ( nIterCount % ( nColorCount - 1 ) ) + 1 ]
				@ nMaxRow - nRow, nCol say "" color saColors[ ( nIterCount % ( nColorCount - 1 ) ) + 1 ]
			else
				// did not diverge -- member of set
				@ nRow, nCol say "" color saColors[ nColorCount ]
				@ nMaxRow - nRow, nCol say "" color saColors[ nColorCount ]
			endif
		next
	next

	@ 24, 0 say padc( "Done.  Press any key to continue", maxCol() + 1 ) color "N/W"
	inkey( 0 )

RETURN


// Function    : IterCount()
// Description : Determine the number of iterations before divergence.
// Parameters  :
//   nReal : real component of complex number z
//   nImag : imaginary component of complex number z
//
//  Details    :
//    This function calculates the iterations of
//    the formula z -> z^2 + c.  The bailout point
//    for the calculation is determined by
//    MAX_DIST_SQR, the square of the distance of
//    point z from the origin (0,0).

STATIC FUNCTION IterCount( nReal, nImag )

	local nIterCount
	local nRealBase := nReal
	local nImagBase := nImag
	local nRealTemp := nReal

	for nIterCount := 1 to MAX_ITERATIONS

		// check for divergence
		if ( ( nReal * nReal ) + ( nImag * nImag ) ;
			> MAX_DIST_SQR )
			// diverges
			exit
		endif

		// calculate next iteration
		nRealTemp := ( ( nReal * nReal ) - ;
			( nImag * nImag ) ) + nRealBase

		nImag     := ( 2 * nReal * nImag ) + nImagBase
		nReal     := nRealTemp

	next

RETURN( nIterCount )

// EOF: Fract.prg
