// File        : Fract2.prg
// Description :
//   This file demonstrates a simple algorithm to
//   calculate and display the Mandelbrot set in text mode.
//
// Copyright   : Wynn Systems Development Group, Ltd.

// Function/Procedure Prototype Table  -  Last Update: 04-20-93 @ 05:05:34pm
// 
// Return Value         Function/Arguments
//   
// aFract[ FR_X_RAT...  STATIC FUNCTION Col2Real( aFract, nCol )
// nIterCount           STATIC FUNCTION IterCount( nReal, nImag )
// aFract[ FR_Y_RAT...  STATIC FUNCTION Row2Imag( aFract, nRow )
// Void                 PROCEDURE DrawNext( aFract )
// Void                 PROCEDURE FractDisp( aWind )
// Void                 PROCEDURE FractProc( aMsg )
// Void                 PROCEDURE FractRestart( aWind )
// Void                 PROCEDURE PauseUpdate( aFract )
// Void                 PROCEDURE UnPauseUpdate( aFract )
// Void                 PROCEDURE fractHide( aWind )
// Void                 PROCEDURE fractMove( aWind, nDir )
// aWind                FUNCTION FractNew( cTitle, nT, nL, nB, nR )
// aFract[ FR_PAUSED ]  FUNCTION isPaused( aFract )

#include "msg.ch"
#include "wind.ch"
#include "inkey.ch"
#include "misc.ch"

#define MAX_DIST_SQR 16
#define MAX_ITERATIONS 12

#define FR_PAUSED    1
#define FR_WINDOW    2
#define FR_X_ORIGIN  3
#define FR_Y_ORIGIN  4
#define FR_X_RATIO   5
#define FR_Y_RATIO   6
#define FR_TITLE     7
#define FR_INCLRS    8
#define FR_CLR_COUNT 9
#define FR_OUTCLR    FR_CLR_COUNT                // one and the same

#define FR_ASIZE     9

#define FR_X_SCALE  6                             // range -3..3
#define FR_Y_SCALE  5                             // range -2.5i..2.5i


FUNCTION FractNew( cTitle, nT, nL, nB, nR, aColors )

	// create new fractal pseudo-object
	LOCAL aFract := ARRAY( FR_ASIZE )
	LOCAL aWind

	aWind := wNew( cTitle, nT, nL, nB, nR, { |aMsg| FractProc( aMsg ) }, aFract )

	aFract[ FR_WINDOW ]   := aWind
	aFract[ FR_PAUSED ]   := FALSE
	aFract[ FR_X_ORIGIN ] := ( nL + nR ) / 2
	aFract[ FR_Y_ORIGIN ] := ( nT + 2 + nB ) / 2
	aFract[ FR_X_RATIO ]  := FR_X_SCALE / ( nL - nR )
	aFract[ FR_Y_RATIO ]  := FR_Y_SCALE / ( nB - nT )
	aFract[ FR_TITLE ]    := cTitle

	IF aColors == NIL
		aFract[ FR_INCLRS ] := ;
			{ "R+/W", "RB/W", "B/W", "BG/W", "G/W", ;
			"R/W", "RB+/W", "BG+/W", "GR+/W", "W+/W", "G+/W", "W/W" }
	ELSE
		aFract[ FR_INCLRS ] := aColors
	ENDIF

	aFract[ FR_OUTCLR ]   := LEN( aFract[ FR_INCLRS ] )

	// assign a constant timer request
	SetTimer( aWind, 0 )

	// reset the cursor column
	aWind[ W_CURS_COL ] := 0

RETURN( aWind )


PROCEDURE FractDisp( aWind )

	LOCAL aFract := aWind[ W_CARGO ]

	// adjust title
	DO CASE
	CASE ( aWind[ W_T ] + aWind[ W_CURS_ROW ] >= aFract[ FR_Y_ORIGIN ] )
		aWind[ W_TITLE ] := aFract[ FR_TITLE ] + " [ Complete ]"
	CASE aFract[ FR_PAUSED ]
		aWind[ W_TITLE ] := aFract[ FR_TITLE ] + " [ Suspended ]"
	OTHERWISE
		aWind[ W_TITLE ] := aFract[ FR_TITLE ] + " [ Drawing ]"
	ENDCASE

	wDisp( aWind )

RETURN


PROCEDURE FractRestart( aWind )

	DISPBEGIN()

	// clear window
	wScroll( aWind, aWind[ W_T ], aWind[ W_L ], aWind[ W_B ], aWind[ W_R ] )

	// reset row and column
	aWind[ W_CURS_ROW ] := 1
	aWind[ W_CURS_COL ] := 0

	// re-establish the timer
	SetTimer( aWind, 0 )

	wHide( aWind )

	// redisplay the fractal
	FractDisp( aWind )

	DISPEND()

RETURN


PROCEDURE fractMove( aWind, nDir )

	LOCAL aFract := aWind[ W_CARGO ]

	wMove( aWind, nDir )
	aFract[ FR_X_ORIGIN ] := ( aWind[ W_L ] + aWind[ W_R ] ) / 2
	aFract[ FR_Y_ORIGIN ] := ( aWind[ W_T ] + aWind[ W_B ] + 2 ) / 2

RETURN


PROCEDURE fractHide( aWind )

	wHide( aWind )

RETURN


PROCEDURE PauseUpdate( aFract )

	LOCAL aWind  := aFract[ FR_WINDOW ]

	aFract[ FR_PAUSED ] := .T.
	aWind[ W_TITLE ] := aFract[ FR_TITLE ] + " [ Paused ]"
	// wTitleDisp( aWind )

RETURN


PROCEDURE UnPauseUpdate( aFract )

	LOCAL aWind  := aFract[ FR_WINDOW ]

	aFract[ FR_PAUSED ] := .F.
	aWind[ W_TITLE ] := aFract[ FR_TITLE ] + " [ Drawing ]"
	// wTitleDisp( aWind )

RETURN


FUNCTION isPaused( aFract )
RETURN( aFract[ FR_PAUSED ] )


PROCEDURE DrawNext( aFract )

	LOCAL aWind   := aFract[ FR_WINDOW ]
	LOCAL nRow    := aWind[ W_CURS_ROW ]           // relative
	LOCAL nCol    := aWind[ W_CURS_COL ]           // relative
	LOCAL nIterCount
	LOCAL nOldRow := ROW()
	LOCAL nOldCol := COL()
	LOCAL cColor

	IF ( ! ( ( aWind[ W_T ] + aWind[ W_CURS_ROW ] >= aFract[ FR_Y_ORIGIN ] ) .or. ;
			isPaused( aFract ) ) )

		// Check for membership in the Mandelbrot set
		IF ( ;
				( nIterCount := IterCount( ;
						Col2Real( aFract, aFract[ FR_X_ORIGIN ] - ( nCol + aWind[ W_L ] ) ), ;
						Row2Imag( aFract, aFract[ FR_Y_ORIGIN ] - ( nRow + aWind[ W_T ] ) ) ;
					) ;
				) ;
			< MAX_ITERATIONS )

			// diverged, display it in a color determined
			// by number of iterations before divergence
			cColor := aFract[ FR_INCLRS ][ ( nIterCount % ( aFract[ FR_CLR_COUNT ] - 1 ) ) + 1 ]
		ELSE
			// did not diverge -- member of set
			cColor := aFract[ FR_OUTCLR ]
		ENDIF

		// display results
		wSay( aWind, nRow, nCol, "", cColor )
		wSay( aWind, aWind[ W_B ] - aWind[ W_T ] - nRow + 1, nCol, "", cColor )

		// advance window cursor
		IF ( ++aWind[ W_CURS_COL ] > aWind[ W_R ] - aWind[ W_L ] )
			aWind[ W_CURS_COL ] := 0
			aWind[ W_CURS_ROW ]++

			IF ( aWind[ W_T ] + aWind[ W_CURS_ROW ] >= aFract[ FR_Y_ORIGIN ] )
				aWind[ W_TITLE ] := aFract[ FR_TITLE ] + " [ Complete ]"

				// kill the timer - no further updates required
				KillTimer( aWind )

				// redisplay title
				wTitleDisp( aWind )
				// TONE( 200, 1 ) ; TONE( 400, 1 ) ; TONE( 800, 1 )
			ENDIF

		ENDIF

		DEVPOS( nOldRow, nOldCol )

	ENDIF

RETURN


// Function    : IterCount()
// Description : Determine the number of iterations before divergence.
// Parameters  :
//   nReal : real component of complex number z
//   nImag : imaginary component of complex number z
//
//  Details    :
//    This function calculates the iterations of
//    the formula z -> z^2 + c.  The bailout point
//    for the calculation is determined by
//    MAX_DIST_SQR, the square of the distance of
//    point z from the origin (0,0).

STATIC FUNCTION IterCount( nReal, nImag )

	LOCAL nIterCount
	LOCAL nRealBase := nReal
	LOCAL nImagBase := nImag
	LOCAL nRealTemp := nReal

	FOR nIterCount := 1 TO MAX_ITERATIONS

		// check for divergence
		IF ( ( nReal * nReal ) + ( nImag * nImag ) ;
			> MAX_DIST_SQR )
			// diverges
			EXIT
		ENDIF

		// calculate next iteration
		nRealTemp := ( ( nReal * nReal ) - ;
			( nImag * nImag ) ) + nRealBase

		nImag     := ( 2 * nReal * nImag ) + nImagBase
		nReal     := nRealTemp

	NEXT

RETURN( nIterCount )


STATIC FUNCTION Col2Real( aFract, nCol )
RETURN( aFract[ FR_X_RATIO ] * nCol )


STATIC FUNCTION Row2Imag( aFract, nRow )
RETURN( aFract[ FR_Y_RATIO ] * nRow )


PROCEDURE FractProc( aMsg )

	LOCAL aWind  := aMsg[ M_WINDOW ]
	LOCAL msg    := aMsg[ M_MSG ]
	LOCAL nParm  := aMsg[ M_PARM ]
	LOCAL aFract := aWind[ W_CARGO ]

	DO CASE
	CASE ( msg == WM_TIMER )
		DrawNext( aFract )

	CASE ( msg == WM_PAINT )
		FractDisp( aWind )

	CASE ( msg == WM_SETFOCUS )
		wGetFocus( aWind )

	CASE ( msg == WM_KILLFOCUS )
		wKillFocus( aWind )

	CASE ( msg == WM_MOVE )
		fractMove( aWind, nParm )

	CASE ( msg == WM_HIDE )
		fractHide( aWind )

	CASE ( msg == WM_DESTROY )
		wKill( aWind )

	CASE ( msg == WM_CHAR )

		DO CASE
		CASE ( nParm == K_SPACE )

			IF ( isPaused( aFract ) )
				UnPauseUpdate( aFract )
			ELSE
				PauseUpdate( aFract )
			ENDIF

		CASE ( nParm == K_ENTER )
			FractRestart( aWind )

		ENDCASE

	ENDCASE

	aWind[ W_LASTTIME ] := seconds()

RETURN

// EOF: Fract2.prg
